@@ -28,10 +28,13 @@ Example
 """
 
 from __future__ import annotations
+
+import dataclasses
 import math
 import sys
 import re as _re
 from typing import Tuple, List, NamedTuple, Union as U
+from functools import lru_cache as _cache
 
 
 _EPS = sys.float_info.epsilon
@@ -42,6 +45,30 @@ number_t = U[int, float]
 _flats  = ["C", "Db", "D", "Eb", "E", "F", "Gb", "G", "Ab", "A", "Bb", "B", "C"]
 _sharps = ["C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B", "C"]
 
+_pitch_class = {
+    'C': 0,
+    'C#': 1,
+    'Db': 1,
+    'D': 2,
+    'D#': 3,
+    'Eb': 3,
+    'E': 4,
+    'E#': 5,
+    'Fb': 4,
+    'F': 5,
+    'F#': 6,
+    'Gb': 6,
+    'G': 7,
+    'G#': 8,
+    'Ab': 8,
+    'A': 9,
+    'A#': 10,
+    'Bb': 10,
+    'B': 11,
+    'B#': 0,
+    'Cb': 11,
+}
+
 _notes2 = {"c": 0, "d": 2, "e": 4, "f": 5, "g": 7, "a": 9, "b": 11}
 
 _r1 = _re.compile(r"(?P<pch>[A-Ha-h][b|#]?)(?P<oct>[-]?[\d]+)(?P<micro>[-+><↓↑][\d]*)?")
@@ -50,21 +77,26 @@ _r2 = _re.compile(r"(?P<oct>[-]?\d+)(?P<pch>[A-Ha-h][b|#]?)(?P<micro>[-+><↓↑
 
 class NoteParts(NamedTuple):
     """
+    Represents the parts of a note, returned by parse_midinote
+
     Attributes:
         octave (int): octave number, 4=central octave
-        noteName (str: "C", "D", "E", ... (diatonic step)
+        diatonic_name (str: "C", "D", "E", ... (diatonic step)
         alteration (str): the alteration as str, "#", "b", "+", "-", ">", "<"
-        centsDeviation (int): number of cents deviation from the chromatic pitch
+        cents_deviation (int): number of cents deviation from the chromatic pitch
     """
     octave: int
-    noteName: str
+    diatonic_name: str
     alteration: str
-    centsDeviation: int
+    cents_deviation: int
 
     @property
-    def alterationCents(self) -> int:
+    def alteration_cents(self) -> int:
         return alteration_to_cents(self.alteration)
 
+    @property
+    def diatonic_step(self) -> int:
+        return 'ABCDEFGAB'.index(self.diatonic_name)
 
 
 class ParsedMidinote(NamedTuple):
@@ -81,27 +113,142 @@ class ParsedMidinote(NamedTuple):
     chromatic_pitch: str
 
 
-class NotatedPitch(NamedTuple):
+_cents_repr_eighthtones = {
+    0: '',
+    25: '>',
+    50: '+',
+    -25: '<',
+    -50: '-'
+}
+
+_cents_repr_quartertones = {
+    0: '',
+    50: '+',
+    -50: '-'
+}
+
+
+def cents_repr(cents:int, eighthToneShortcuts=True) -> str:
     """
+    Return the string representation of cents
+
+    ======   =======================
+    cents     string representation
+    ======   =======================
+    0          ''
+    15         +15
+    25
+    """
+    if eighthToneShortcuts:
+        shortcut = _cents_repr_eighthtones.get(cents)
+    else:
+        shortcut = _cents_repr_quartertones.get(cents)
+    if shortcut:
+        return shortcut
+    else:
+        return f'+{cents}' if cents > 0 else str(cents)
+
+
+@dataclasses.dataclass
+class NotatedPitch:
+    """
+    A NotatedPitch is a parsed notename which can be queried in relation
+    to its musical notation
+
     Attributes:
         octave: the octave (4=central octave)
         diatonic_index: 0=C, 1=D, 2=E, ...
-        diatonic_step: "C", "D", "E", ...
+        diatonic_name: "C", "D", "E", ...
         chromatic_index: 0=C, 1=C#, 2=D, ...
-        chromatic_step: "C", "Db", ...
-        diatonic_alteration: for C# this would be 1.0, for Db this would be -1.0
+        chromatic_name: "C", "Db", ...
+        diatonic_alteration: the alteration in relation to the diatonic pitch
+            For C# this would be 1.0, for Db this would be -1.0
         chromatic_alteration: for C#+50 this would be 0.5
         accidental_name: the name of the accidental used ('natural', 'natural-up','quarter-sharp', etc.)
     """
     octave: int
     diatonic_index: int
-    diatonic_step: str
+    diatonic_name: str
     chromatic_index: int
-    chromatic_step: str
+    chromatic_name: str
     diatonic_alteration: float
     chromatic_alteration: float
+    "docs for chromatic alteration"
+
     accidental_name: str
 
+    @property
+    def fullname(self) -> str:
+        return f"{self.octave}{self.chromatic_name}{self.cents_str}"
+
+    @property
+    def vertical_position(self) -> int:
+        """
+        Abstract value indicating the vertical notated position 
+        """
+        return self.octave * 7 + self.diatonic_index
+
+    @property
+    def is_white_key(self) -> bool:
+        return self.diatonic_alteration == 0
+
+    @property
+    def is_black_key(self) -> bool:
+        return not(self.is_white_key)
+
+    @property
+    def cents_deviation(self) -> int:
+        """The cents deviation from the notated chromatic pitch"""
+        return int(self.chromatic_alteration * 100)
+
+    @property
+    def cents_sign(self) -> str:
+        if self.chromatic_alteration == 0:
+            return ''
+        elif self.chromatic_alteration > 0:
+            return '+'
+        else:
+            return '-'
+
+    @property
+    def cents_str(self) -> str:
+        cents = self.cents_deviation
+        if cents == 0:
+            return ''
+        elif cents == 50:
+            return '+'
+        elif cents == -50:
+            return '-'
+        elif cents > 0:
+            return f'+{cents}'
+        else:
+            return str(cents)
+
+    def alteration_direction(self, min_alteration=0.5):
+        """
+        Returns the direction of the alteration
+
+        (the table assumes min_alteration==0.5)
+
+        ===== ====================
+        Note  Alteration Direction
+        ===== ====================
+        4C     0
+        4C#    1
+        4Eb   -1
+        4C+    1
+        4F-25  0
+        4Bb    -1
+        4A+25  0
+        ===== ====================
+
+        """
+        if self.diatonic_alteration >= min_alteration:
+            return 1
+        elif self.diatonic_alteration <= -min_alteration:
+            return -1
+        return 0
+
 
 class Converter:
     """
@@ -230,7 +377,7 @@ class Converter:
 
     def pianofreqs(self, start="A0", stop="C8") -> List[float]:
         """
-        Generate an array of the frequencies representing all the piano keys
+        Generate an array of the frequencies for all the piano keys
 
         Args:
             start: the starting note
@@ -314,8 +461,8 @@ class Converter:
             midinote: the midinote to analyze
 
         Returns:
-            a NoteParts instance, a named tuple with the fields: `octave`, `noteName`,
-            `alteracion` and `centsDeviation`
+            a NoteParts instance, a named tuple with the fields: `octave`, `notename`,
+            `alteracion` and `cents_deviation`
 
         """
         i = int(midinote)
@@ -369,6 +516,46 @@ class Converter:
         """
         return self.m2n(self.n2m(notename))
 
+    def as_midinotes(self, x: U[List[pitch_t], List[str], str, float]) -> List[float]:
+        """
+        Tries to interpret `x` as a list of pitches, returns these as midinotes
+
+        Args:
+            x: either list of midinotes (floats/ints), a list of notenames (str), one
+                str with notenames (divided by spaces), or a single notename or midinote
+
+        Returns:
+            the corresponding list of midinotes.
+
+        Example
+        ~~~~~~~
+
+            >>> as_midinotes(["4G", "4C"])
+            [67., 60.]
+            >>> as_midinotes((67, 60))
+            [67., 60.]
+            >>> as_midinotes("4G 4C 4C+10hz")
+            [67., 60., 60.65]
+
+        """
+        if isinstance(x, str):
+            notenames = x.split()
+            midinotes = [self.str2midi(n) for n in notenames]
+        elif isinstance(x, (list, tuple)):
+            midinotes = []
+            for n in x:
+                if isinstance(n, str):
+                    midinotes.append(self.str2midi(n))
+                elif isinstance(n, (int, float)):
+                    midinotes.append(n)
+                else:
+                    raise TypeError(misc.type_error_msg(n, str, int, float))
+        elif isinstance(x, (float, int)):
+            midinotes = [x]
+        else:
+            raise TypeError(misc.type_error_msg(x, str, 'list[str]', 'list[float]'))
+        return midinotes
+
 
 def n2m(note: str) -> float:
     """
@@ -461,6 +648,14 @@ def n2m(note: str) -> float:
     return (octave + 1) * 12 + pc + micro
 
 
+def is_valid_notename(notename: str, minpitch=12) -> bool:
+    try:
+        midi = n2m(notename)
+        return midi >= minpitch
+    except ValueError:
+        return False
+
+
 def _pitchname(pitchidx: int, micro: float) -> str:
     """
     Given a pitchindex (0-11) and a microtonal alteracion (between -0.5 and +0.5),
@@ -767,7 +962,7 @@ def _parse_centstr(centstr: str) -> int:
     return cents
 
 
-def split_notename(notename: str) -> Tuple[int, str, str, int]:
+def split_notename(notename: str) -> NoteParts:
     """
     Return (octave, letter, alteration (#, b), cents)
 
@@ -801,7 +996,10 @@ def split_notename(notename: str) -> Tuple[int, str, str, int]:
             octave = int(notename[1])
             cursor = 2
         centstr = notename[cursor:]
-        cents = _parse_centstr(centstr)
+        try:
+            cents = _parse_centstr(centstr)
+        except ValueError:
+            raise ValueError(f"Could not parse cents '{centstr}' while parsing note '{notename}'")
     else:
         # 4C#-10
         octave = int(notename[0])
@@ -820,7 +1018,7 @@ def split_notename(notename: str) -> Tuple[int, str, str, int]:
             else:
                 centstr = rest
             cents = _parse_centstr(centstr)
-    return octave, letter.upper(), alter, cents
+    return NoteParts(octave, letter.upper(), alter, cents)
 
 
 def split_cents(notename: str) -> Tuple[str, int]:
@@ -844,76 +1042,93 @@ def split_cents(notename: str) -> Tuple[str, int]:
     ========   ============
 
     """
-    octave, letter, alter, cents = split_notename(notename)
-    alterchar = "b" if alter == -1 else "#" if alter == 1 else ""
-    return str(octave) + letter + alterchar, cents
-
+    parts = split_notename(notename)
+    return f"{parts.octave}{parts.diatonic_name}{parts.alteration}", parts.cents_deviation
+    
 
 def enharmonic(notename: str) -> str:
     """
-    Return the enharmonic version of notename. 
-
-    Notes without an alteration have no enharmonic 
-    (double alterations are never used)
+    Returns the enharmonic variant of notename
 
+    For simplicity we considere a possible enharmonic variant a note
+    with the same sounding pitch and an alteration smaller than 150
+    cents from the note without any alteration (no double sharps or
+    flats). Also not accepted are enharmonics like Fb or E#
 
     Args:
-        notename: the notename to find an enharmonic to
+        notename (str): the note to find an enharmonic variant to
 
     Returns:
-        the enharmonic version of the note (this can be the
-        same as the original)
-
-    ==========     ============
-    original       enharmonic
-    ==========     ============
-    4C+50          4Db-50
-    4A+25          unchanged
-    4G#+25         4Ab+25
-    4Eb-25         4D#-25
-    4G+30          unchanged
-    ==========     ============
+        either the enharmonic variant or the note itself
 
+    =====  ===========  ================
+    Note   Enharmonic   Has Enharmonic?
+    =====  ===========  ================
+    4#     4Db           x
+    4C+    4Db-          x
+    4E     4E            -
+    4E#    4F            x
+    4A+10  4A+10         -
+    4E-25  4E-25         -
+    4E-    4D#+          x
     """
-    midinote = n2m(notename)
-    diatonicsteps = "CDEFGAB"
-    if int(midinote) == midinote:
-        return notename
-    octave, letter, alteration, cents = split_notename(notename)
-    sign = "+" if cents > 0 else "-" if cents < 0 else ""
-    pitchidx = diatonicsteps.index(letter)
-    if alteration != 0:
-        # a black key
-        if alteration == 1:
-            # turn sharp to flat
-            basenote = diatonicsteps[pitchidx+1] + "b"
-            return f"{octave}{basenote}{sign}{abs(cents)}"
-        else:
-            # turn flat into sharp
-            basenote = diatonicsteps[pitchidx-1] + "#"
-            return f"{octave}{basenote}{sign}{abs(cents)}"
-    else:
-        if cents == 50:
-            # 4D+50 -> 4Eb-50
-            # 4B+50 -> 5C-50
-            if letter == "B":
-                return f"{octave+1}C-"
-            basenote = diatonicsteps[pitchidx+1] + "b"
-            return f"{octave}{basenote}-"
-        elif cents == -50:
-            # 4D-50 -> 4C#+50
-            # 4C-50 -> 3B+50
-            if letter == "C":
-                return f"{octave-1}B+"
-            basenote = diatonicsteps[pitchidx+1]+"b"
-            return f"{octave}{basenote}-"
-        else:
+    p = notated_pitch(notename)
+    if abs(p.diatonic_alteration) < 1:
+        if abs(p.cents_deviation) < 50:
             return notename
+        if p.cents_deviation >= 50:
+            chrom = _flats[p.chromatic_index+1]
+            octave = p.octave if p.chromatic_name != "B" else p.octave+1
+            return f"{octave}{chrom}{cents_repr(p.cents_deviation-100)}"
+        else:
+            # 4E- : 4D#+
+            # 4E-60 : 4D#+40
+            chrom = _sharps[(p.chromatic_index-1)%12]
+            octave = p.octave if p.chromatic_name != "C" else p.octave-1
+            return f"{octave}{chrom}{cents_repr(100+p.cents_deviation)}"
+    if p.diatonic_alteration >= 1:
+        # 4C# : 4Db
+        # 4C#+25 : 4Db+25
+        # 4C#+ : 4D-
+        # 4C#+60 : 4D-40
+        # 4D#-25 : 4Eb-25
+        # 4D#-70 : 4D+30
+        if 0 <= abs(p.cents_deviation) < 50:
+            chrom = _flats[p.chromatic_index]
+            return f"{p.octave}{chrom}{p.cents_str}"
+        elif 50 <= p.cents_deviation < 100:
+            chrom = _flats[p.chromatic_index+1]
+            centstr = cents_repr(p.cents_deviation-100)
+        elif -100 < p.cents_deviation < 50:
+            chrom = _flats[(p.chromatic_index-1)%12]
+            centstr= cents_repr(100+p.cents_deviation)
+        else:
+            raise ValueError("???", p)
+        return f"{p.octave}{chrom}{centstr}"
+    else: #  p.diatonic_alteration == -1:
+        # 4Db : 4C#
+        # 4Db-25 : 4C#-25
+        # 4Db-   : 4C+
+        # 4Db-60 : 4C+40
+        # 4Eb+25 : 4D#+25
+        # 4Eb+70 : 4E-30
+        if 0 <= abs(p.cents_deviation) < 50:
+            chrom = _sharps[p.chromatic_index]
+            return f"{p.octave}{chrom}{p.cents_str}"
+        elif 50 <= p.cents_deviation < 100:
+            chrom = _sharps[p.chromatic_index-1]
+            centstr = cents_repr(p.cents_deviation-100)
+        elif -100 < p.cents_deviation:
+            chrom = _sharps[p.chromatic_index-1]
+            centstr = cents_repr(100+p.cents_deviation)
+        else:
+            raise ValueError("???", p)
+        return f"{p.octave}{chrom}{centstr}"
 
 
 def pitch_round(midinote: float, semitoneDivisions=4) -> Tuple[str,int]:
     """
-    Round midinote to the next (possibly microtonal) pitch
+    Round midinote to the next (possibly microtonal) note
 
     Returns the rounded notename and the cents deviation
     from the original pitch to the next semitone
@@ -993,12 +1208,12 @@ _centsToAccidentalName = {
 }
 
 
-def accidental_name(alterationCents: int, semitoneDivisions=4) -> str:
+def accidental_name(alteration_cents: int, semitoneDivisions=4) -> str:
     """
     The name of the accidental corresponding to the given cents
 
     Args:
-        alterationCents: 100 = sharp, -50 = quarter-flat, etc.
+        alteration_cents: 100 = sharp, -50 = quarter-flat, etc.
         semitoneDivisions: number of divisions of the semitone
 
     Returns:
@@ -1024,29 +1239,64 @@ def accidental_name(alterationCents: int, semitoneDivisions=4) -> str:
             -150    three-quarters-flat
     """
     assert semitoneDivisions in {1, 2, 4}, "semitoneDivisions should be 1, 2, or 4"
-    if alterationCents < -150 or alterationCents > 150:
-        raise ValueError(f"alterationCents should be between -150 and 150, "
-                         f"got {alterationCents}")
     centsResolution = 100 // semitoneDivisions
-    alterationCents = round(alterationCents / centsResolution) * centsResolution
-    return _centsToAccidentalName[alterationCents]
+    alteration_cents = round(alteration_cents / centsResolution) * centsResolution
+    return _centsToAccidentalName[alteration_cents]
 
 
 def _roundres(x:float, resolution:float) -> float:
     return round(x/resolution)*resolution
 
 
-def notated_pitch(midinote: float, divsPerSemitone=4) -> NotatedPitch:
+def vertical_position(note: str) -> int:
     """
-    Convert a (fractional) midinote to a NotatedPitch
+    Return the vertical notated position of a note
+
+    The only relevant information for the vertical position
+    is the octave and the diatonic pitch class. So, 4G# and 4G
+    have the same vertical position, 4Ab and 4G# do not (the vertical
+    position of 4Ab is 4*7+6=34, for 4G# it is 33)
+    """
+    notated = notated_pitch(note)
+    return notated.vertical_position()
+
+
+@_cache(maxsize=0)
+def notated_pitch(pitch: U[float, str], divsPerSemitone=4) -> NotatedPitch:
+    """
+    Convert a note or a (fractional) midinote to a NotatedPitch
 
     Args:
-        midinote: a midinote as float (60=4C)
-        divsPerSemitone: number of divisions per semitone
+        pitch: a midinote as float (60=4C), or a notename
+        divsPerSemitone: number of divisions per semitone (only relevant
+            when passing a midinote as pitch
 
     Returns:
         the corresponding pitch as NotatedPitch
     """
+    
+    if isinstance(pitch, (int, float)):
+        return _notated_pitch_midinote(pitch, divsPerSemitone)
+    return _notated_pitch_notename(pitch)
+
+
+def _notated_pitch_notename(notename: str) -> NotatedPitch:
+    parts = split_notename(notename)
+    diatonic_index = 'CDEFGABC'.index(parts.diatonic_name)
+    chromatic_note = parts.diatonic_name + parts.alteration
+    cents = parts.cents_deviation
+    diatonic_alteration = (alteration_to_cents(parts.alteration)+cents) / 100
+    return NotatedPitch(octave=parts.octave,
+                        diatonic_index=diatonic_index,
+                        diatonic_name=parts.diatonic_name,
+                        chromatic_index=_pitch_class[chromatic_note],
+                        chromatic_name=chromatic_note,
+                        diatonic_alteration=diatonic_alteration,
+                        chromatic_alteration=cents/100,
+                        accidental_name=accidental_name(int(diatonic_alteration*100)))
+
+
+def _notated_pitch_midinote(midinote: float, divsPerSemitone=4) -> NotatedPitch:
     rounded_midinote = _roundres(midinote, 1/divsPerSemitone)
     parsed_midinote = parse_midinote(rounded_midinote)
     notename = m2n(rounded_midinote)
@@ -1061,9 +1311,9 @@ def notated_pitch(midinote: float, divsPerSemitone=4) -> NotatedPitch:
 
     return NotatedPitch(octave=octave,
                         diatonic_index=diatonic_index,
-                        diatonic_step=letter,
+                        diatonic_name=letter,
                         chromatic_index=parsed_midinote.pitchindex,
-                        chromatic_step=chromaticStep,
+                        chromatic_name=chromaticStep,
                         diatonic_alteration=diatonicAlteration,
                         chromatic_alteration=cents/100,
                         accidental_name=accidental_name(int(diatonicAlteration*100)))
@@ -1083,6 +1333,7 @@ f2n = _converter.f2n
 freqround = _converter.freqround
 normalize_notename = _converter.normalize_notename
 str2midi = _converter.str2midi
+as_midinotes = _converter.as_midinotes
 
 
 # --- Amplitude converters ---
@@ -1112,3 +1363,21 @@ def amp2db(amp: float) -> float:
     amp = max(amp, _EPS)
     return math.log10(amp) * 20
 
+
+def _test_enharmonic():
+    assert (result := enharmonic("4E+")) == "4F-", f"got {result}"
+    assert (result := enharmonic("4F#")) == "4Gb", f"got {result}"
+    assert (result := enharmonic("4C+")) == "4Db-", f"got {result}"
+    assert (result := enharmonic("4G-")) == "4F#+", f"got {result}"
+    assert (result := enharmonic("4G+60")) == "4Ab-40", f"got {result}"
+    assert (result := enharmonic("4E+25")) == "4E+25", f"got {result}"
+    assert (result := enharmonic("5Eb-55")) == "5D+45", f"got {result}"
+    assert (result := enharmonic("4G-25")) == "4G-25", f"got {result}"
+    assert (result := enharmonic("4C-25")) == "4C-25", f"got {result}"
+    assert (result := enharmonic("4C-")) == "3B+", f"got {result}"
+    assert (result := enharmonic("3B-")) == "3A#+", f"got {result}"
+    assert (result := enharmonic("4F-55")) == "4E+45", f"got {result}"
+
+
+def _tests():
+    _test_enharmonic()
\ No newline at end of file
